/******************************************************************************/
/*									      */
/*			    SAMPLE SUPPORT FOR RAINE			      */
/*									      */
/*			  modified by Hiromitsu Shioya			      */
/*			  change SEAL library 09/20/98			      */
/*									      */
/******************************************************************************/
// Avoid to use this thing in linux. I can be wrong, but apparently there is
// no way to decide wether to use signed or unsigned samples.
// And seal seems to be obliged to convert the samples in linux to match its
// default format. Too bad, it works really badly with looping voices...

#include <time.h>
#include "audio.h"
#include "raine.h"
#include "sasound.h"

#include "games.h"

#include <sys/time.h>

#define SEAL

int RaineSoundCardTotal,
  RaineSoundCard;

int STREAM_BUFFER_MAXA,
  STREAM_BUFFER_MAXB,
  MODEB_UPDATE_COUNT,
  MODEB_FRAME_SIZE,
  MODEB_MASK,
  MODEB_ERROR_MAX;

int max_mixer_volume = 0; // unused for now

// uncomment USE_MODE if you want to choose streaming mode.
// Totally untested with seal
//#define USE_MODE 1

// USE_8BITS is something from the old days of seal. Don't know if it works
//#define USE_8BITS 1

#ifdef USE_MODE
//int sound_stream_mode = SOUND_STREAM_NORMAL; // SOUND_STREAM_WAIT;
int sound_stream_mode = SOUND_STREAM_WAIT;
#endif

#define print_debug printf

/*

dump channels (debug)

*/

//#define DUMP_CHANNELS

#ifdef DUMP_CHANNELS
FILE *stream_out[NUMVOICES];
#endif

/*

mame seal interface. syncs works better, but it is much slower, blame
seal or your pc for this. the sample rate seal streams play at is nearly
always different from the one it *should* be running at.

*/


/*
 *  Volume Section
 */

#define VOLUME_MIN	(0)			// Minimum volume
#define VOLUME_MAX	(255)			// Maximum volume

/*
 *  Panning Section
 */

#define PAN_LEFT	(0)			// Left panning
#define PAN_CENTRE	(128)			// Central panning
#define PAN_RIGHT	(255)			// Right panning

/*
 *  Sample Control Section
 */

#define CPU_FPS 	(60)			// Frames per second of emulated hardware

/*
 *  Other stuff
 */

#define UPDATE_60_FPS				// Turn off the counting stuff. Emulation
						// speed stays more constant this way.

static int samp_modeb_timer;

SoundRec      *SndMachine = NULL, snd_entry;

/* audio related stuff */

HAC	    hVoice[NUMVOICES];
LPAUDIOWAVE lpWave[NUMVOICES];
int	    Volumi[NUMVOICES];
int	    MasterVolume;
int	    playing[NUMVOICES];
int	    vchan[NUMVOICES];
int	    ventry[NUMVOICES];
int	    vbover_err, vbunder_err;
int	    *vpan[NUMVOICES];

int	    SampleRate;
int	    audio_sample_rate;
int	    sound_emu_count;

static int reserved_channel = 0;

int sound_stream_mode = SOUND_STREAM_WAIT;

static int stream_buffer_max;

static int pause_sound;

void saCheckPlayStream( void );

/*******************************************************************************************/
/**** osd_??? functions 								****/
/*******************************************************************************************/
int get_play_channels( int request )
{
  return saGetPlayChannels( request );
}

void reset_play_channels(void)
{
  saResetPlayChannels();
}

#ifdef USE_8BITS
void osd_play_streamed_sample(int channel,signed char *data,int len,int freq,int volume, int pan){
#if USE_MODE
  if( !sound_stream_mode )  saPlayStreamedSampleBase( channel, (signed char *)data, len, freq, volume, 8, pan );
  else
#endif
    saPlayBufferedStreamedSampleBase( channel, (signed char *)data, len, freq, volume, 8, pan );
}
#endif

void osd_play_streamed_sample_16(int channel,signed short *data,int len,int freq,int volume, int pan){
#if USE_MODE
  if( !sound_stream_mode )  saPlayStreamedSampleBase( channel, (signed char *)data, len, freq, volume, 16, pan );
  else
#endif
    saPlayBufferedStreamedSampleBase( channel, (signed char *)data, len, freq, volume, 16, pan );
}

void osd_ym2203_update(void)
{
}

/*******************************************************************************************/
/*  streams v35x									   */
/*******************************************************************************************/

#include "streams.c"

/*******************************************************************************************/
/*  sa???Sound										   */
/*******************************************************************************************/
/******************************************/
/*    update sound			  */
/******************************************/
void saUpdateSound( int nowclock )
{
   if( ! GameSound ) return;
   if( ! RaineSoundCard ) return;
   if( ! audio_sample_rate ) return;
   if( ! SndMachine ) return;

   if( nowclock ){

      streams_sh_update();
      AUpdateAudio();

   }
   else{
 // This part was called by profile.c, normally exactly 60 times / sec.
#if USE_MODE
      if( sound_stream_mode ){
#ifdef RAINE_DOS
		  AUpdateAudio();
#endif
		  saCheckPlayStream();
      }
#endif

   }

}

/******************************************************************************/
/*									      */
/*			    SOUND CARD INFORMATION			      */
/*									      */
/******************************************************************************/

static char *sound_name_list[16][60];
static int sound_id_list[16];

char *sound_card_name( int num )
{
   AUDIOINFO audio_info;
   AUDIOCAPS audio_caps;
   if (sound_name_list[1][0] == 0) { // init list
     int n;
     int id=0;
     AInitialize();
     for (n=0; n<0x200; n++) {
       audio_info.nDeviceId = n;
       audio_info.wFormat = AUDIO_FORMAT_16BITS | AUDIO_FORMAT_STEREO;
       audio_info.nSampleRate = audio_sample_rate;
       if (!AGetAudioDevCaps( audio_info.nDeviceId, &audio_caps )) {
	 sound_id_list[id] = n;
	 strncpy((char*)sound_name_list[id++],audio_caps.szProductName,60);
       }
     }
     RaineSoundCardTotal = id;
   }

   if (num>= 0 && num < RaineSoundCardTotal) {
     return (char*)sound_name_list[num];
   } else return "???";
}

int sound_card_id( int i )
{
  if (i >= 0 && i < RaineSoundCardTotal)
    return sound_id_list[i];
  return 0;
}

/******************************************/
/*    setup sound			  */
/******************************************/
BOOL saInitSoundCard( int soundcard, int sample_rate )
{
   AUDIOINFO audio_info;
   AUDIOCAPS audio_caps;

   int i;

   soundcard = sound_card_id(soundcard);

   if( AInitialize() != AUDIO_ERROR_NONE ){
      #ifdef HACKS
      print_debug("saInitSoundCard(): Seal failed to initialize\n");
      #endif
      audio_sample_rate = 0;
      return FALSE;
   }

   if( soundcard <= 0 ){		// no soundcard (not going to autodetect)
      #ifdef HACKS
      print_debug("saInitSoundCard(): OK [No Card]\n");
      #endif
      audio_sample_rate = 0;
      return TRUE;
   }

   // Set card id

   audio_info.nDeviceId = soundcard;

   // 16 bit only

   audio_info.wFormat = AUDIO_FORMAT_16BITS | AUDIO_FORMAT_STEREO;

   // Seal is crap, it clicks, it's slow, it's bugged (why are we ever using it?)

   if(audio_sample_rate>44098)
   audio_sample_rate = 44098;

   // Sample Rate

   audio_info.nSampleRate = audio_sample_rate;

   // Open audio

   if( AOpenAudio(&audio_info) != AUDIO_ERROR_NONE ){
      #ifdef HACKS
      print_debug("saInitSoundCard(): Audio Initialization Failed!\n");
      #endif
      audio_sample_rate = 0;
      return FALSE;
   }

   // Find out what seal chose for us

   AGetAudioDevCaps( audio_info.nDeviceId, &audio_caps );
#ifdef HACKS
   print_debug(
      "Sound - '%s' %d-bit %s %u Hz\n",
      audio_caps.szProductName,
      audio_info.wFormat & AUDIO_FORMAT_16BITS ? 16 : 8,
      audio_info.wFormat & AUDIO_FORMAT_STEREO ? "stereo" : "mono",
      audio_info.nSampleRate
   );
#endif

   // Update sample rate (incase seal chose a different one)
/*
   if(audio_info.nSampleRate != audio_sample_rate)
      audio_sample_rate = audio_info.nSampleRate;
*/
   // Open and allocate voices, allocate waveforms

   if( AOpenVoices(NUMVOICES) != AUDIO_ERROR_NONE ){
      #ifdef HACKS
      print_debug("saInitSoundCard(): Voice Initialization Failed\n");
      #endif
      audio_sample_rate = 0;
      return FALSE;
   }

   for( i = 0; i < NUMVOICES; i++ ){
      if( ACreateAudioVoice(&hVoice[i]) != AUDIO_ERROR_NONE ){
	  #ifdef HACKS
	  print_debug("saInitSoundCard(): Voice %d Creation Failed\n",i);
	  #endif
	  audio_sample_rate = 0;
	  return FALSE;
      }
      ASetVoicePanning( hVoice[i], PAN_CENTRE );

      lpWave[i]  = 0;
      playing[i] = 0;

      #ifdef DUMP_CHANNELS
      sprintf(s,"stream%02d.raw",i);
      stream_out[i] = fopen(s,"wb");
      #endif
   }

   #ifdef RAINE_DOS
   AUpdateAudio();
   #endif

   #ifdef HACKS
   print_debug("saInitSoundCard(): SEAL initialize OK\n");
   #endif

   /**** stream buffer mode ****/
#if USE_MODE
   if( !sound_stream_mode )  stream_buffer_max = STREAM_BUFFER_MAXA;
   else
#endif
     stream_buffer_max = STREAM_BUFFER_MAXB;
   samp_modeb_timer = 0;

   MasterVolume = 100;
   //reserved_channel = 0;

   if(!init_sound_emulators())
      return FALSE;

   return TRUE;
}

/******************************************/
/*    setup sound			  */
/******************************************/

void init_sound_list(void)
{
   SOUND_INFO *sound_src;
   int ta;

   sound_src = current_game->sound_list;

   if(sound_src){

   saStopSoundEmulators();

   for( ta = 0; ta < SND_CONTROL_MAX; ta++ ){
      SndMachine->init[ta] = SOUND_NONE;
      SndMachine->intf[ta] = NULL;
   }

   SndMachine->first = 0;
   SndMachine->control_max = 0;

   ta = 0;

   while(sound_src[ta].interface){

   SndMachine->init[ta] = sound_src[ta].type;
   SndMachine->intf[ta] = sound_src[ta].interface;

   ta++;

   }

   SndMachine->control_max = ta;

   GameSound = 1;

   }
}

/******************************************/
/*    destroy sound			  */
/******************************************/

void saDestroyChannel( int chan )
{
   if( lpWave[chan] ){
      AStopVoice( hVoice[chan] );
      ADestroyAudioData( lpWave[chan] );
      free( lpWave[chan] );
      free( vpan[chan] );
      lpWave[chan] = 0;
      playing[chan] = 0;
      vpan[chan] = 0;
   }
}

void saDestroySound( int remove_all_resources )
{
   int i;

   #ifdef HACKS
   print_debug("saDestroySound: Removing SEAL\n");
   #endif

   // pause_raine_ym3812();

   if(remove_all_resources) saStopSoundEmulators();

   for( i = 0; i < NUMVOICES; i++ ){
      saDestroyChannel( i );
      ADestroyAudioVoice( hVoice[i] );

      #ifdef DUMP_CHANNELS
      fclose(stream_out[i]);
      #endif

   }
   ACloseVoices();
   ACloseAudio();
   //SndMachine = NULL;
   #ifdef HACKS
   print_debug("saDestroySound: OK\n");
   #endif
}

void sa_pause_sound(void)
{
   int i;

   pause_sound	    = 1;
   samp_modeb_timer = 0;
   vbover_err	    = 0;
   vbunder_err	    = 0;

   // pause_raine_ym3812();
#if 1
   // There is a remarquable bug in the win32 version of seal : voices can't
   // be stopped !!!
   saDestroySound(0);
#else
   for( i = 0; i < NUMVOICES; i++ ){
     if (playing[i])
       AStopVoice( hVoice[i] );
   }
#endif
}

void sa_unpause_sound(void)
{
  int i;
  pause_sound	    = 0;
#if 1
  saInitSoundCard(RaineSoundCard,audio_sample_rate);
#else
  for( i = 0; i < NUMVOICES; i++ ){
    if (playing[i])
      AStartVoice(hVoice[i]);
  }
#endif
}

/******************************************/
/*    buffer check			  */
/******************************************/
void saCheckPlayStream( void )
{
#if 1
  return;
  /* This function was trying to resynchronize streams before.
     It did not do it the right way.
     It's useless now !!! */
#else
   int i, len, rlen;
   DWORD pos;
   int vp, ve;

   if( !pause_sound ){

      samp_modeb_timer = (samp_modeb_timer+1) % MODEB_FRAME_SIZE;	// <60> res = [0..59]

      vp = samp_modeb_timer / MODEB_UPDATE_COUNT;			// <3>	res = [0..19]

      for( i = 0; i < NUMVOICES; i++ ){ 				// ***** SCAN VOICES

	 if( playing[i] ){						// ***** PROCESS ENABLED VOICE

	    ve = ventry[i] % MODEB_MASK;				// <60/3=20> res = [0..19]

	    if( vp == ve ){						// ***** PROCESS ENABLED STREAM VOICE

	       /**** buffer update check ****/
	       AGetVoicePosition( hVoice[i], &pos );
	       len = (lpWave[i]->dwLength / stream_buffer_max) * MODEB_UPDATE_COUNT;	// length of <3> buffers
	       rlen = len * vp; 							// buffer point where sample should go

	       if( (pos < rlen) || (pos >= (rlen+len)) ){		// Make sure it is not playing this buffer right now

		  /**** check play buffer/write buffer position ****/
		  ve = (ventry[i] + 1) * MODEB_UPDATE_COUNT;		//

		  /*if( (vchan[i] - ve) > 0 ){*/
		  /*if( (vchan[i] - ve) > -MODEB_UPDATE_COUNT ){*/

		  if( ((vchan[i] - ve) > (0-MODEB_UPDATE_COUNT)) && (vbunder_err < (MODEB_ERROR_MAX*2))){

		     if( (vchan[i] - ve) < 0 )	vbunder_err++;
		     else{
			if( vbunder_err >= 0 )	vbunder_err--;
		     }
		     /**** all buffer update (not overrun) ****/
		     for( i = 0; i < NUMVOICES; i++ ){
			if( playing[i] ){
			   vbover_err = 0;
			   ASetVoicePanning( hVoice[i], (UINT)*(vpan[i] + samp_modeb_timer) );
			   ventry[i]++;

			   #ifdef DUMP_CHANNELS
			   //len  = (lpWave[i]->dwLength / stream_buffer_max) * MODEB_UPDATE_COUNT;
			   //rlen = len * vp;
			   //fwrite( (((UINT8 *) (lpWave[i]->lpData))+rlen), 1, len, stream_out[i]);
			   if(vp==0) fwrite( lpWave[i]->lpData, 1, lpWave[i]->dwLength, stream_out[i]);
			   #endif

			}
		     }

		  }
		  else{

		     #ifdef HACKS
		     print_ingame(120,"Now buf-p[over] %08x %08x", vchan[i], ve );
		     #endif
		     /**** all buffer update (over run) ****/
		     for( i = 0; i < NUMVOICES; i++ ){
			if( playing[i] ){
			   //AStopVoice( hVoice[i] );
			   vbunder_err = 0;
			   vbover_err = 0;
			   len	= (lpWave[i]->dwLength / stream_buffer_max) * MODEB_UPDATE_COUNT;
			   rlen = len * vp;
			   //AWriteAudioData( lpWave[i], rlen, len );
			   //APlayVoice( hVoice[i], lpWave[i] );
			   ASetVoicePanning( hVoice[i], (UINT)*(vpan[i] + samp_modeb_timer) );
			   vchan[i] = (ventry[i] + 1) * MODEB_UPDATE_COUNT;
			   ventry[i]++;
			}
		     }

		  }
		  return;		// Completed
	       }
	       else{			// ERROR: Required buffer was being played

		  vbover_err++; 	/* error count */
		  print_ingame(120,"ve:%08x, vc:%08x, verr:%08x", ventry[0]*MODEB_UPDATE_COUNT, vchan[0], vbover_err );
		  if( vbover_err >= MODEB_ERROR_MAX ){
		     /**** all buffer restart ****/
		     vbover_err       = 0;
		     vbunder_err      = 0;
		     samp_modeb_timer = 0;

		     for( i = 0; i < NUMVOICES; i++ ){
			#if 0
			ventry[i] = 1;
			vchan[i] = MODEB_UPDATE_COUNT;
			#else
			saDestroyChannel( i );
			#endif
		     }

		     #ifdef HACKS
		     print_ingame(60,"Buffer stop: Restarting streams.");
		     #endif
		  }
		  return;		// Completed

	       }
	       /**** pos check end ****/
	    }			// ***** END PROCESS ENABLED STREAM VOICE
	 }			// ***** END PROCESS ENABLED VOICE
      } 			// ***** END SCAN VOICES
   }				// ***** END PLAY SOUND
#endif
}

/*******************************************************************************************/
/*******************************************************************************************/
/******************************************/
/*    play samples			  */
/******************************************/
void saPlayBufferedStreamedSampleBase( int channel, signed char *data, int len, int freq, int volume, int bits , int pan ){
  int i;
  if( audio_sample_rate == 0 || channel >= NUMVOICES )	return;
  if( SndMachine == NULL )  return;
  if( !playing[channel] ){
    if( lpWave[channel] ){
      AStopVoice( hVoice[channel] );
      ADestroyAudioData( lpWave[channel] );
      free( lpWave[channel] );
      lpWave[channel] = 0;
    }
    if( (lpWave[channel] = (LPAUDIOWAVE)malloc(sizeof(AUDIOWAVE))) == 0 )  return;
    if( (vpan[channel] = (int *)malloc(sizeof(int)*stream_buffer_max)) == NULL )  return;

    lpWave[channel]->wFormat	 = (bits == 8 ? AUDIO_FORMAT_8BITS :
    AUDIO_FORMAT_16BITS) | AUDIO_FORMAT_MONO | AUDIO_FORMAT_LOOP;
    lpWave[channel]->nSampleRate = audio_sample_rate;
    lpWave[channel]->dwLength	 = stream_buffer_max*len;
    lpWave[channel]->dwLoopStart = 0;
    lpWave[channel]->dwLoopEnd	 = stream_buffer_max*len;
    if( ACreateAudioData(lpWave[channel]) != AUDIO_ERROR_NONE ){
      free( lpWave[channel] );
      free( vpan[channel] );
      lpWave[channel] = 0;
      vpan[channel] = 0;
      return;
    }
    memset( lpWave[channel]->lpData, 0, stream_buffer_max*len );
    ASetVoicePosition(hVoice[channel],0);
    APlayVoice( hVoice[channel], lpWave[channel] );
    ASetVoiceFrequency( hVoice[channel], freq );
    ASetVoiceVolume( hVoice[channel], MasterVolume*volume/400 );
    ASetVoicePanning( hVoice[channel], (UINT)pan );
    playing[channel] = 1;	/* use front surface */
    /**** make sound temp. buffer ****/
    vchan[channel] = MODEB_UPDATE_COUNT;
    memcpy( &lpWave[channel]->lpData[len * MODEB_UPDATE_COUNT], data, len );
    for( i = 0; i <= MODEB_UPDATE_COUNT; i++ ){
      *(vpan[channel] + i) = pan;
    }
    vchan[channel]++;
    ventry[channel] = 1;
  } else{
    int s_pos = vchan[channel]%stream_buffer_max;
    memcpy( &lpWave[channel]->lpData[len * s_pos], data, len );
    *(vpan[channel] + s_pos) = pan;
    vchan[channel]++;
    if (s_pos == MODEB_UPDATE_COUNT) {
      ASetVoicePosition(hVoice[channel],0);
      // APlayVoice( hVoice[channel], lpWave[channel] );
    }
  }
  Volumi[channel] = volume/4;
}

#ifdef USE_8BITS
/******************************************/
/*    play samples			  */
/******************************************/
void saPlayStreamedSampleBase( int channel, signed char *data, int len, int freq, int volume, int bits , int pan ){
  if( audio_sample_rate == 0 || channel >= NUMVOICES )	return;
  if( SndMachine == NULL )  return;
  if( !playing[channel] ){
    if( lpWave[channel] ){
      AStopVoice( hVoice[channel] );
      ADestroyAudioData( lpWave[channel] );
      free( lpWave[channel] );
      lpWave[channel] = 0;
    }
    if( (lpWave[channel] = (LPAUDIOWAVE)malloc(sizeof(AUDIOWAVE))) == 0 )  return;

    lpWave[channel]->wFormat = (bits == 8 ? AUDIO_FORMAT_8BITS :
    AUDIO_FORMAT_16BITS) | AUDIO_FORMAT_MONO | AUDIO_FORMAT_LOOP;
    lpWave[channel]->nSampleRate = audio_sample_rate;
    lpWave[channel]->dwLength = stream_buffer_max*len;
    lpWave[channel]->dwLoopStart = 0;
    lpWave[channel]->dwLoopEnd = stream_buffer_max*len;
    if( ACreateAudioData(lpWave[channel]) != AUDIO_ERROR_NONE ){
      free( lpWave[channel] );
      lpWave[channel] = 0;
      return;
    }
    memset(lpWave[channel]->lpData,0,stream_buffer_max*len);
    memcpy(lpWave[channel]->lpData,data,len);
    /* upload the data to the audio DRAM local memory */
    AWriteAudioData(lpWave[channel],0,stream_buffer_max*len);
    APlayVoice(hVoice[channel],lpWave[channel]);
    ASetVoiceFrequency(hVoice[channel],freq);
    ASetVoiceVolume(hVoice[channel],MasterVolume*volume/400);
    ASetVoicePanning(hVoice[channel], (UINT)pan );
    playing[channel] = 1;	/* use front surface */
    vchan[channel] = 1;
    ventry[channel] = 1;

    //print_debug("first stream entry. [%d:%d:%d:%d]\n", channel, len, freq, volume );
  } else{
    DWORD pos;
#if 0
    int nownum = vchan[channel]%STREAM_BUFFER_MAXA;
    for(;;){
#ifdef RAINE_DOS
      AUpdateAudio();
#endif
      AGetVoicePosition(hVoice[channel],&pos);

      if( !nownum && pos >= len )  break;
      if( nownum == (stream_buffer_max-1) ){
	if( pos < (len*nownum) )  break;
      } else{
	if( (pos < len*(nownum) || pos >= (len*(nownum+1))) )  break;
      }
      /*
      if ((vchan[channel]%STREAM_BUFFER_MAXA) == 0 && pos >= len) break;
      if ((vchan[channel]%STREAM_BUFFER_MAXA) == 1 && (pos < len || pos >= (2*len))) break;
      if ((vchan[channel]%STREAM_BUFFER_MAXA) == 2 && pos < (2*len)) break;
      */
    }
#else
    AGetVoicePosition(hVoice[channel],&pos);
#ifdef RAINE_DOS
    AUpdateAudio();
#endif
#endif

    memcpy( &lpWave[channel]->lpData[len * (vchan[channel]%STREAM_BUFFER_MAXA)], data, len );
    AWriteAudioData( lpWave[channel], len*(vchan[channel]%STREAM_BUFFER_MAXA), len );
    vchan[channel]++;
    //print_debug("next stream entry. [%d:%d:%d:%d::%d]\n", channel, len, freq, volume, pos );
  }
  Volumi[channel] = volume/4;
}
#endif

/******************************************************************************/
/*									      */
/*			  SOUND CHANNEL ALLOCATION			      */
/*									      */
/******************************************************************************/

int saGetPlayChannels( int request )
{
   int ret_value = reserved_channel;
   reserved_channel += request;
   return ret_value;
}

void saResetPlayChannels( void )
{
   reserved_channel = 0;
}

/******************************************************************************/
/*									      */
/*			  VOLUME / PANNING CONTROL			      */
/*									      */
/******************************************************************************/

void saInitVolPan( void )
{
   int i;

   for( i = 0; i < MAX_STREAM_CHANNELS; i++ ){
      SampleVol[i] = VOLUME_MAX;
      SamplePan[i] = PAN_CENTRE;
   }
}

void saSetVolume( int num, int data )
{
   SampleVol[num] = data <= VOLUME_MAX ? data : VOLUME_MAX;
}

void saSetPan( int num, int data )
{
   SamplePan[num] = data;
}

void saSetPanMulti( int num, int data )
{
   int	i;

   for( i = 0; i < stream_joined_channels[num]; i++ )
      SamplePan[num + i] = data;
}

/******************************* END OF FILE **********************************/

